import numpy as np
import os
import cv2
from tqdm import tqdm
import torch

from loader import test_lightfield_loader, image_check
from testfunctions import image_shot, image_reconstruct, image_concat
from evalation import calc_psnr, evalation
from model import*
from saver import SaveMasks

def test(args, dstdir, save_log):

        with torch.no_grad():
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

            frame_num = args.end_frame - args.start_frame+1
            view_point_num = args.viewpoint**2
            print("frame num : %d"%frame_num)

            if args.para:
                    model_num = args.shots_num
                    load_image_num = args.shots_num*2-1
            else:
                    model_num = 1
                    load_image_num = args.shots_num

            #get image size
            fname = os.path.join(args.test_data, "%04d"%args.start_frame, "00_00.png")
            size_h, size_w, chanel_num = image_check(fname, args)
            print("image size : H x W  =  %d x %d"%(size_h, size_w))
            print("chanel num : %d"%chanel_num)
            save_log.log("image size : H x W = %d x %d\n"%(size_h, size_w))
            save_log.log("chanel : %d\n"%(chanel_num))
            save_log.log("\n")

            #output direcrories
            save_mask_dir = os.path.join(dstdir, "Mask")
            save_reconstructedimg_dir = os.path.join(dstdir, args.test_data, "ReconstructedImage")
            save_acquiredimg_dir = os.path.join(dstdir, args.test_data, "AcquiredImage")

            os.makedirs(save_mask_dir, exist_ok=True)
            os.makedirs(save_reconstructedimg_dir, exist_ok=True)
            os.makedirs(save_acquiredimg_dir, exist_ok=True)

            #Load trained models
            aperture_mask = []
            pixel_mask = []
            PSreconstruct = []
            vdsr = []
            for s in range(args.shots_num):
                    tmp_mask = ApertureMask()
                    tmp_mask.load_state_dict(torch.load("%s/ApertureMask%01d.pth"%(args.model_data, s+1)))
                    aperture_mask.append(tmp_mask.to(device))
                    aperture_mask[s].cn01.bias.data[:] = 0.0


            for m in range(model_num):
                    tmp_model = PixelSuffuleReconstruct()
                    tmp_model.load_state_dict(torch.load("%s/PSreconstruct%01d.pth"%(args.model_data, m+1)))
                    PSreconstruct.append(tmp_model.to(device))

                    tmp_model = VeryDeepSuperResolution()
                    tmp_model.load_state_dict(torch.load("%s/vdsr%01d.pth"%(args.model_data, m+1)))
                    vdsr.append(tmp_model.to(device))

            for n in range(args.pix_mask_size**2):
                    tmp_mask = PixelMask()
                    tmp_mask.load_state_dict(torch.load("%s/ForwardPixelMask/ForwardPixelMask%01d.pth"%(args.model_data, n+1)))
                    pixel_mask.append(tmp_mask.to(device))
                    pixel_mask[n].cn01.bias.data[:] = 0.0

            mse_sum = 0
            cnt = 0

            reconstructed_light_field = np.zeros((load_image_num, view_point_num, size_h, size_w, chanel_num), np.float32)

            #Test start
            print("start simulation")
            for frame in range(args.start_frame, args.end_frame-load_image_num+1, args.shots_num):

                    for s in range(args.shots_num):
                            os.makedirs(os.path.join(save_reconstructedimg_dir, "%04d"%(frame+s)), exist_ok=True)

                    lf_images = np.zeros([load_image_num, chanel_num, view_point_num, size_h, size_w], np.float32)
                    lf_images = test_lightfield_loader(frame, args, load_image_num)
                    lf_images = torch.from_numpy(lf_images.astype(np.float32)).clone().to(device)

                    acquired_image = np.zeros((size_h, size_w, chanel_num), np.float32)

                    # aquire image & reconstruct each chanel
                    for ch in range(chanel_num):

                            gray_sub_acquired_images, gray_acquired_image = image_shot(lf_images[:,ch:ch+1, :, :, :], aperture_mask, pixel_mask, args)
                            gray_cut_reconstruct_light_field = image_reconstruct(gray_sub_acquired_images, PSreconstruct, vdsr, args)

                            reconstructed_light_field[:,:,:,:,ch] = image_concat(gray_cut_reconstruct_light_field, reconstructed_light_field[:,:,:,:,ch])
                            acquired_image[:,:,ch] = gray_acquired_image.to("cpu").detach().numpy()

                    reconstructed_light_field[reconstructed_light_field < 0.0]  = 0.0
                    reconstructed_light_field[reconstructed_light_field > 1.0]  = 1.0


                    for s in range(args.shots_num):
                            #Save reconstructed light field
                            tmp_light_field = np.zeros((view_point_num, size_h, size_w, chanel_num), np.uint8)
                            for t in range(view_point_num):
                                    tmp_light_field[t,:,:,:] = (reconstructed_light_field[s,t,:,:,:]*255.0).astype(np.uint8)
                                    fname = os.path.join(save_reconstructedimg_dir,
                                                         "%04d/%02d_%02d.png"%(frame+s, t//args.viewpoint, t%args.viewpoint))
                                    cv2.imwrite(fname, tmp_light_field[t,:,:,:])

                            if frame == args.start_frame and s < args.shots_num-1:
                                    continue
                            else:
                                    # Calculate MSE PSNR
                                    gt_light_field = lf_images[s].to("cpu").detach().numpy().copy()
                                    gt_light_field = np.transpose(gt_light_field, (1,2,3,0)).astype(np.uint8)
                                    mse, psnr= evalation(gt_light_field, tmp_light_field, args)
                                    ave_mse = np.average(mse)
                                    ave_psnr = calc_psnr(ave_mse)
                                    mse_sum += ave_mse
                                    cnt += 1
                                    save_log.log("Frame %04d \tmse : %f \tpsnr : %f\n"%(frame+s, ave_mse, ave_psnr))
                                    print("Frame %04d \tmse : %f \tpsnr : %f"%(frame+s, ave_mse, ave_psnr))

                    #Save acquired images
                    fname = os.path.join(save_acquiredimg_dir, "%04d-%04d.png"%(frame, frame+args.shots_num-1))
                    cv2.imwrite(fname, (acquired_image*255/(args.shots_num*view_point_num)).astype(np.uint8))

                    for s in range(args.shots_num-1):
                            reconstructed_light_field[s,:,:,:,:] = reconstructed_light_field[s+args.shots_num,:,:,:,:]

            for s in range(args.shots_num-1):
                    tmp_light_field = np.zeros((view_point_num, size_h, size_w, chanel_num), np.uint8)
                    os.makedirs(os.path.join(save_reconstructedimg_dir, "%04d"%(args.end_frame-args.shots_num+s+1)), exist_ok=True)
                    for t in range(view_point_num):
                            tmp_light_field[t,:,:,:] = (reconstructed_light_field[s,t,:,:,:]*255.0).astype(np.uint8)
                            fname = os.path.join(save_reconstructedimg_dir,
                                                 "%04d/%02d_%02d.png"%(args.end_frame-args.shots_num+s+1, t//args.viewpoint, t%args.viewpoint))
                            cv2.imwrite(fname,tmp_light_field[t,:,:,:])

            total_ave_mse = mse_sum/cnt
            total_ave_psnr = calc_psnr(total_ave_mse)
            print("\nTotal Average ::    mse : %f    psnr : %f\n"%(total_ave_mse, total_ave_psnr))
            save_log.log("\nTotal Average ::    mse : %f    psnr : %f\n"%(total_ave_mse, total_ave_psnr))

            #Save mask images
            save_mask = SaveMasks(savedir = save_mask_dir)
            save_mask.save_masks(aperture_mask, pixel_mask)
